# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2001-2006  Donald N. Allingham
# Copyright (C) 2008       Gary Burton
# Copyright (C) 2010       Nick Hall
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Base view for Place Views
"""

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gdk
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.lib import Place
from gramps.gui.views.listview import ListView, TEXT, ICON
from gramps.gen.errors import WindowActiveError
from gramps.gui.views.bookmarks import PlaceBookmarks
from gramps.gen.config import config
from gramps.gui.dialog import ErrorDialog
from gramps.gui.pluginmanager import GuiPluginManager
from gramps.gui.ddtargets import DdTargets
from gramps.gui.editors import EditPlace
from gramps.gui.filters.sidebar import PlaceSidebarFilter
from gramps.gui.merge import MergePlace
from gramps.gen.plug import CATEGORY_QR_PLACE
from gramps.gen.utils.location import located_in
from gramps.gui.uimanager import ActionGroup

# -------------------------------------------------------------------------
#
# internationalization
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext


# -------------------------------------------------------------------------
#
# PlaceBaseView
#
# -------------------------------------------------------------------------
class PlaceBaseView(ListView):
    """base view class for place views, be they flat list or tree"""

    COL_NAME = 0
    COL_ID = 1
    COL_TITLE = 2
    COL_TYPE = 3
    COL_CODE = 4
    COL_LAT = 5
    COL_LON = 6
    COL_PRIV = 7
    COL_TAGS = 8
    COL_CHAN = 9
    COL_SEARCH = 11
    # column definitions
    COLUMNS = [
        (_("Name"), TEXT, None),
        (_("ID"), TEXT, None),
        (_("Title"), TEXT, None),
        (_("Type"), TEXT, None),
        (_("Code"), TEXT, None),
        (_("Latitude"), TEXT, None),
        (_("Longitude"), TEXT, None),
        (_("Private"), ICON, "gramps-lock"),
        (_("Tags"), TEXT, None),
        (_("Last Changed"), TEXT, None),
    ]
    # default setting with visible columns, order of the col, and their size
    CONFIGSETTINGS = (
        ("columns.visible", [COL_NAME, COL_ID, COL_TYPE, COL_CODE]),
        (
            "columns.rank",
            [
                COL_NAME,
                COL_TITLE,
                COL_ID,
                COL_TYPE,
                COL_CODE,
                COL_LAT,
                COL_LON,
                COL_PRIV,
                COL_TAGS,
                COL_CHAN,
            ],
        ),
        ("columns.size", [250, 250, 75, 100, 100, 150, 150, 40, 100, 100]),
    )
    ADD_MSG = _("Add a new place")
    EDIT_MSG = _("Edit the selected place")
    DEL_MSG = _("Delete the selected place")
    MERGE_MSG = _("Merge the selected places")
    FILTER_TYPE = "Place"
    QR_CATEGORY = CATEGORY_QR_PLACE

    def __init__(self, pdata, dbstate, uistate, title, model, nav_group):
        signal_map = {
            "place-add": self.row_add,
            "place-update": self.row_update,
            "place-delete": self.row_delete,
            "place-rebuild": self.object_build,
        }

        self.mapservice = config.get("interface.mapservice")
        self.mapservicedata = {}
        self.map_action_group = None

        ListView.__init__(
            self,
            title,
            pdata,
            dbstate,
            uistate,
            model,
            signal_map,
            PlaceBookmarks,
            nav_group,
            multiple=True,
            filter_class=PlaceSidebarFilter,
        )

        uistate.connect("placeformat-changed", self.build_tree)

        _ui = self.__create_maps_menu_actions()
        self.additional_uis.append(_ui)

    def navigation_type(self):
        return "Place"

    def setup_filter(self):
        """Build the default filters and add them to the filter menu.
        This overrides the listview method because we use the hidden
        COL_SEARCH that has alt names as well as primary name for name
        searching"""
        self.search_bar.setup_filter(
            [
                (
                    self.COLUMNS[pair[1]][0],
                    self.COL_SEARCH if pair[1] == self.COL_NAME else pair[1],
                    pair[1] in self.exact_search(),
                )
                for pair in self.column_order()
                if pair[0]
            ]
        )

    def define_actions(self):
        ListView.define_actions(self)
        self._add_action("GotoMap", self.gotomap)

    def change_page(self):
        """
        Called by viewmanager at end of realization when arriving on the page
        At this point the Toolbar is created. We need to:
          1. get the menutoolbutton
          2. add all possible map services in the drop down menu
          3. add the actions that correspond to clicking in this drop down menu
          4. set icon and label of the menutoolbutton now that it is realized
          5. store label so it can be changed when selection changes
        """
        ListView.change_page(self)

    def __create_maps_menu_actions(self):
        """
        Function creating a menu and actions that are used as dropdown menu
        from the menutoolbutton
        """
        _bar = """
            <item>
              <attribute name="action">win.MapChoice</attribute>
              <attribute name="target">%s</attribute>
              <attribute name="label">%s</attribute>
            </item>
            """
        menu = ""

        # select the map services to show
        self.mapservicedata = {}
        servlist = GuiPluginManager.get_instance().get_reg_mapservices()
        for pdata in servlist:
            key = pdata.id.replace(" ", "-")
            menu += _bar % (key, pdata.name)
            self.mapservicedata[key] = pdata

        if not self.mapservicedata:
            return self.additional_ui
        if self.mapservice not in self.mapservicedata:
            # stored val no longer exists, use the most recent key instead
            self.set_mapservice(None, key)

        self._add_toggle_action("MapChoice", self.set_mapservice, "", self.mapservice)

        label = self.mapservice_label()
        _ui = self.additional_ui[:]
        _ui.append(self.map_ui_menu % menu)
        _ui.append(self.map_ui % label)
        return _ui

    def set_mapservice(self, action, value):
        """
        change the service that runs on click of the menutoolbutton
        used as callback menu on menu clicks
        """
        if action:
            action.set_state(value)
            self.mapservice = value.get_string()
        else:
            self.mapservice = value
        config.set("interface.mapservice", self.mapservice)
        config.save()
        _ui = self.__create_maps_menu_actions()
        self.uimanager.add_ui_from_string(_ui)
        self.uimanager.update_menu()
        return False

    def mapservice_label(self):
        """
        return the current label for the menutoolbutton
        """
        return self.mapservicedata[self.mapservice].name

    def gotomap(self, *obj):
        """
        Run the map service
        """
        # First test if any map service is available
        if not len(self.mapservicedata):
            msg = _("No map service is available.")
            msg2 = _("Check your installation.")
            ErrorDialog(msg, msg2, parent=self.uistate.window)
            return

        place_handles = self.selected_handles()
        try:
            place_handle = self.selected_handles()[0]
        except IndexError:
            msg = _("No place selected.")
            msg2 = _(
                "You need to select a place to be able to view it"
                " on a map. Some Map Services might support multiple"
                " selections."
            )
            ErrorDialog(msg, msg2, parent=self.uistate.window)
            return

        # TODO: support for descriptions in some cases. For now, pass None
        # TODO: Later this might be 'Birth of William' ....
        places = [(x, None) for x in place_handles]

        # run the mapservice:
        pmgr = GuiPluginManager.get_instance()
        serv = self.mapservicedata[self.mapservice]
        mod = pmgr.load_plugin(serv)
        if mod:
            servfunc = getattr(mod, serv.mapservice)
            servfunc()(self.dbstate.db, places, self.uistate)
        else:
            print("Failed to load map plugin, see Plugin Manager")

    def drag_info(self):
        return DdTargets.PLACE_LINK

    def get_stock(self):
        return "gramps-place"

    #
    # Defines the UI string for UIManager
    #
    additional_ui = [
        """
      <placeholder id="LocalExport">
        <item>
          <attribute name="action">win.ExportTab</attribute>
          <attribute name="label" translatable="yes">Export View...</attribute>
        </item>
      </placeholder>
""",
        """
      <section id="AddEditBook">
        <item>
          <attribute name="action">win.AddBook</attribute>
          <attribute name="label" translatable="yes">_Add Bookmark</attribute>
        </item>
        <item>
          <attribute name="action">win.EditBook</attribute>
          <attribute name="label" translatable="no">%s...</attribute>
        </item>
      </section>
"""
        % _("Organize Bookmarks"),
        """
      <placeholder id="CommonGo">
      <section>
        <item>
          <attribute name="action">win.Back</attribute>
          <attribute name="label" translatable="yes">_Back</attribute>
        </item>
        <item>
          <attribute name="action">win.Forward</attribute>
          <attribute name="label" translatable="yes">_Forward</attribute>
        </item>
      </section>
      </placeholder>
""",
        """
      <section id='CommonEdit' groups='RW'>
        <item>
          <attribute name="action">win.Add</attribute>
          <attribute name="label" translatable="yes">_Add...</attribute>
        </item>
        <item>
          <attribute name="action">win.Edit</attribute>
          <attribute name="label">%s</attribute>
        </item>
        <item>
          <attribute name="action">win.Remove</attribute>
          <attribute name="label" translatable="yes">_Delete</attribute>
        </item>
        <item>
          <attribute name="action">win.Merge</attribute>
          <attribute name="label" translatable="yes">_Merge...</attribute>
        </item>
      </section>
"""
        % _("_Edit...", "action"),  # to use sgettext()
        """
        <placeholder id='otheredit'>
        <item>
          <attribute name="action">win.FilterEdit</attribute>
          <attribute name="label" translatable="yes">"""
        """Place Filter Editor</attribute>
        </item>
        </placeholder>
""",  # Following are the Toolbar items
        """
    <placeholder id='CommonNavigation'>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">go-previous</property>
        <property name="action-name">win.Back</property>
        <property name="tooltip_text" translatable="yes">"""
        """Go to the previous object in the history</property>
        <property name="label" translatable="yes">_Back</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RO'>
      <object class="GtkToolButton">
        <property name="icon-name">go-next</property>
        <property name="action-name">win.Forward</property>
        <property name="tooltip_text" translatable="yes">"""
        """Go to the next object in the history</property>
        <property name="label" translatable="yes">_Forward</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    </placeholder>
""",
        """
    <placeholder id='BarCommonEdit'>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">list-add</property>
        <property name="action-name">win.Add</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">_Add...</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">gtk-edit</property>
        <property name="action-name">win.Edit</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">Edit...</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">list-remove</property>
        <property name="action-name">win.Remove</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">_Delete</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <child groups='RW'>
      <object class="GtkToolButton">
        <property name="icon-name">gramps-merge</property>
        <property name="action-name">win.Merge</property>
        <property name="tooltip_text">%s</property>
        <property name="label" translatable="yes">_Merge...</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
    </child>
    <placeholder id="PlaceMapUi"> </placeholder>
    </placeholder>
"""
        % (ADD_MSG, EDIT_MSG, DEL_MSG, MERGE_MSG),
        """
    <menu id="Popup">
      <section id="PopUpTree">
        <item>
          <attribute name="action">win.Back</attribute>
          <attribute name="label" translatable="yes">_Back</attribute>
        </item>
        <item>
          <attribute name="action">win.Forward</attribute>
          <attribute name="label" translatable="yes">Forward</attribute>
        </item>
      </section>
      <section>
        <item>
          <attribute name="action">win.Add</attribute>
          <attribute name="label" translatable="yes">_Add...</attribute>
        </item>
        <item>
          <attribute name="action">win.Edit</attribute>
          <attribute name="label">%s</attribute>
        </item>
        <item>
          <attribute name="action">win.Remove</attribute>
          <attribute name="label" translatable="yes">_Delete</attribute>
        </item>
        <item>
          <attribute name="action">win.Merge</attribute>
          <attribute name="label" translatable="yes">_Merge...</attribute>
        </item>
      </section>
      <section>
        <placeholder id='QuickReport'>
        </placeholder>
      </section>
      <section>
        <item>
          <attribute name="action">win.GotoMap</attribute>
          <attribute name="label" translatable="yes">"""
        """_Look up with Map Service</attribute>
        </item>
      </section>
    </menu>
"""
        % _("_Edit...", "action"),
    ]  # to use sgettext()

    map_ui_menu = """
      <menu id="MapBtnMenu">
        %s
      </menu>
    """

    map_ui = (
        """<placeholder id="PlaceMapUi">
    <child>
      <object class="GtkToolButton" id="GotoMap">
        <property name="icon-name">go-jump</property>
        <property name="action-name">win.GotoMap</property>
        <property name="tooltip_text" translatable="yes">"""
        """Attempt to see selected locations with a Map Service """
        """(OpenstreetMap, Google Maps, ...)</property>
        <property name="label">%s</property>
        <property name="use-underline">True</property>
      </object>
      <packing>
        <property name="homogeneous">False</property>
      </packing>
     </child>
    <child>
      <object class="GtkToolItem">
        <child>
          <object class="GtkMenuButton">
            <property name="tooltip_text" translatable="yes">"""
        """Select a Map Service</property>
            <property name="menu-model">MapBtnMenu</property>
            <property name="relief">GTK_RELIEF_NONE</property>
            <property name="use-popover">False</property>
          </object>
        </child>
      </object>
    </child>
    </placeholder>
    """
    )

    def add(self, *obj):
        try:
            EditPlace(self.dbstate, self.uistate, [], Place())
        except WindowActiveError:
            pass

    def remove(self, *obj):
        ht_list = []
        for handle in self.selected_handles():
            for _link in self.dbstate.db.find_backlink_handles(handle, ["Place"]):
                msg = _("Cannot delete place.")
                msg2 = _(
                    "This place is currently referenced by another place."
                    " First remove the places it contains."
                )
                ErrorDialog(msg, msg2, parent=self.uistate.window)
                return
            ht_list.append(("Place", handle))
        self.remove_selected_objects(ht_list)

    def edit(self, *obj):
        for handle in self.selected_handles():
            place = self.dbstate.db.get_place_from_handle(handle)
            try:
                EditPlace(self.dbstate, self.uistate, [], place)
            except WindowActiveError:
                pass

    def merge(self, *obj):
        """
        Merge the selected places.
        """
        mlist = self.selected_handles()

        if len(mlist) != 2:
            msg = _("Cannot merge places.")
            msg2 = _(
                "Exactly two places must be selected to perform a merge. "
                "A second place can be selected by holding down the "
                "control key while clicking on the desired place."
            )
            ErrorDialog(msg, msg2, parent=self.uistate.window)
        else:
            if located_in(self.dbstate.db, mlist[0], mlist[1]) or located_in(
                self.dbstate.db, mlist[1], mlist[0]
            ):
                msg = _("Cannot merge places.")
                msg2 = _(
                    "Merging these places would create a cycle in the "
                    "place hierarchy."
                )
                ErrorDialog(msg, msg2, parent=self.uistate.window)
            else:
                MergePlace(
                    self.dbstate, self.uistate, [], mlist[0], mlist[1], self.merged
                )

    def merged(self):
        """
        Rebuild the model after a merge to reflect changes in the hierarchy.
        """
        if not (self.model.get_flags() & Gtk.TreeModelFlags.LIST_ONLY):
            self.build_tree()

    def get_handle_from_gramps_id(self, gid):
        obj = self.dbstate.db.get_place_from_gramps_id(gid)
        if obj:
            return obj.get_handle()
        else:
            return None

    def tag_updated(self, handle_list):
        """
        Update tagged rows when a tag color changes.
        """
        all_links = set([])
        for tag_handle in handle_list:
            links = set(
                [
                    link[1]
                    for link in self.dbstate.db.find_backlink_handles(
                        tag_handle, include_classes="Place"
                    )
                ]
            )
            all_links = all_links.union(links)
        self.row_update(list(all_links))

    def add_tag(self, transaction, place_handle, tag_handle):
        """
        Add the given tag to the given place.
        """
        place = self.dbstate.db.get_place_from_handle(place_handle)
        place.add_tag(tag_handle)
        self.dbstate.db.commit_place(place, transaction)

    def remove_tag(self, transaction, place_handle, tag_handle):
        """
        Remove the given tag from the given place.
        """
        place = self.dbstate.db.get_place_from_handle(place_handle)
        place.remove_tag(tag_handle)
        self.dbstate.db.commit_place(place, transaction)

    def get_default_gramplets(self):
        """
        Define the default gramplets for the sidebar and bottombar.
        """
        return (
            ("Place Filter",),
            (
                "Place Details",
                "Place Enclosed By",
                "Place Encloses",
                "Place Gallery",
                "Place Citations",
                "Place Notes",
                "Place Backlinks",
            ),
        )


def make_callback(func, val):
    return lambda x, y: func(val)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2005-2007  Donald N. Allingham
# Copyright (C) 2008-2009  Gary Burton
# Copyright (C) 2009       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2023       Christopher Horn
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
This package implements access to Gramps configuration.
"""

# ---------------------------------------------------------------
#
# Python modules
#
# ---------------------------------------------------------------
import configparser
import copy
import errno
import logging
import os
import sys
import time

# ---------------------------------------------------------------
#
# Gramps modules
#
# ---------------------------------------------------------------
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


def clean_up(parser):
    # Needed so that the parser can be removed by the garabge collection
    for section in parser:
        delattr(parser._proxies[section], "getboolean")
        delattr(parser._proxies[section], "getint")
        delattr(parser._proxies[section], "getfloat")
        del parser._proxies[section]
    del parser._converters


def safe_eval(exp):
    # restrict eval to empty environment
    return eval(exp, {})


# ---------------------------------------------------------------
#
# ConfigManager
#
# ---------------------------------------------------------------
class ConfigManager:
    """
    Class to construct the singleton CONFIGMAN where all
    settings are stored.
    """

    PLUGINS = {}

    def __init__(self, filename=None, plugins=None):
        """
        Configure manager constructor takes an optional filename, and
        plugin path.

        The data dictionary stores the settings::

           self.data[section][setting] = value

        The value has a type that matches the default. It is an error
        to attempt to set the setting to a different type. To change
        the type, you must re-register the setting, and re-set the
        value. Values can be any simple type in Python (except,
        currently longs, which are saved as ints to avoid type
        errors). This includes: str, int, list, tuple, dict, float,
        etc. Of course, composite types must themselves be composed of
        simple types.

        The default values are given in Python code and stored here
        on start-up::

           self.default[section][setting] = default_value

        Callbacks are stored as callables here::

           self.callbacks[section][setting] = (id, func)

        The default filename (usually the one you are reading from)
        is stored as self.filename. However, you can save to another
        filename using self.save(otherfilename).

        :param filename: (if given) is a fullpath.
        :param plugins: (if given) is a relative path to filename.

        """
        self._cb_id = 0  # callback id counter
        self.config_path, dummy_config_filename = os.path.split(
            os.path.abspath(filename)
        )
        self.filename = filename  # fullpath and filename, or None
        self.plugins = plugins  # relative directory name, or None
        self.callbacks = {}
        self.default = {}
        self.data = {}
        self.reset()

    def register_manager(
        self, name, override="", use_plugins_path=True, use_config_path=False
    ):
        """
        Register a plugin manager.

        :param name: is used as the key of the config manager singleton. It is
                     also be used as the base filename (unless an override is
                     given, or use_config_path or use_plugins_path is True).
        :param override: is used to override the default location of the .ini
                         file.
        :param use_config_path: if True, use this ConfigManager's path as the
                                new manager's path, ignoring any path given in
                                override.

        Override is either:

        - a full path+filename ending in .ini
        - a filename ending in .ini
        - a dir path to put ini file into
        - a full path+filename to get dir to put ini file into
        - a ConfigManager instance

        Examples::

          >>> config.register_manager("Simple", use_plugins_path=False)
          # will use the calling programs directory, and "Simple.ini"
          >>> config.register_manager("Simple", __file__,
                                      use_plugins_path=False)
          # will use the __file__'s directory, and "Simple.ini"
          >>> config.register_manager("Simple", "c:\\temp",
                                      use_plugins_path=False)
          # will use the given directory, "c:\\temp\\Simple.ini"
          >>> config.register_manager("Simple", use_config_path=True)
          # will use config's path: ~/.gramps/gramps32/Simple.ini
          >>> config.register_manager("Simple", "Other.ini")
          # will use config + plugins path: ~/.gramps/gramps32/plugins/Other.ini
          >>> config.register_manager("Simple", "/tmp/Other.ini",
                                      use_plugins_path=False)
          # will use /tmp/Other.ini
        """
        if isinstance(override, str):  # directory or filename
            if override:
                path, ininame = os.path.split(os.path.abspath(override))
            else:
                path, ininame = os.path.split(sys._getframe(1).f_code.co_filename)
            if not ininame.endswith(".ini"):
                ininame = f"{name}.ini"
            if use_config_path:
                path = self.config_path
            elif use_plugins_path:
                path = os.path.join(self.config_path, self.plugins)
            filename = os.path.join(path, ininame)
            plugin = ConfigManager(filename)
        elif isinstance(override, ConfigManager):
            plugin = override
        else:
            raise AttributeError("plugin needs to be a file or ConfigManager")
        ConfigManager.PLUGINS[name] = plugin
        return ConfigManager.PLUGINS[name]

    def get_manager(self, name):
        """
        Return manager for a plugin.
        """
        if name in ConfigManager.PLUGINS:
            return ConfigManager.PLUGINS[name]
        raise AttributeError(f"config '{name}': does not exist")

    def has_manager(self, name):
        """
        Check if have manager for a plugin.
        """
        return name in ConfigManager.PLUGINS

    def init(self):
        """
        Either loads from an existing ini file, or saves to it.
        """
        if self.filename:
            if os.path.exists(self.filename):
                self.load()
            else:
                self.save()

    def __getitem__(self, item):
        """
        For item access, such as config["interface.dont-ask"]
        """
        return self.get(item)

    def __setitem__(self, item, value):
        """
        For item assignment, such as config["interface.dont-ask"] = True
        """
        self.set(item, value)

    def reset(self, key=None):
        """
        Resets one, a section, or all settings values to their defaults.
        This does not disconnect callbacks.
        """
        if key is None:
            section = None
            setting = None
        elif "." in key:
            section, setting = key.split(".", 1)
        else:  # key is not None and doesn't have a "."
            section = key
            setting = None

        # Now, do the reset on the right parts:
        if section is None:
            self.data = {}
            for section in self.default:
                self.data[section] = {}
                for setting in self.default[section]:
                    self.data[section][setting] = copy.deepcopy(
                        self.default[section][setting]
                    )
        elif setting is None:
            self.data[section] = {}
            for setting in self.default[section]:
                self.data[section][setting] = copy.deepcopy(
                    self.default[section][setting]
                )
        else:
            self.data[section][setting] = copy.deepcopy(self.default[section][setting])
        # Callbacks are still connected

    def get_sections(self):
        """
        Return all section names.
        """
        return list(self.data.keys())

    def get_section_settings(self, section):
        """
        Return all section setting names.
        """
        return list(self.data[section].keys())

    def load(self, filename=None, oldstyle=False):
        """
        Loads an .ini into self.data.
        """
        if filename is None:
            filename = self.filename
        if filename and os.path.exists(filename):
            parser = configparser.RawConfigParser()
            try:  # see bugs 5356, 5490, 5591, 5651, 5718, etc.
                parser.read(filename, encoding="utf8")
            except Exception as err:
                msg1 = _(
                    "WARNING: could not parse file:\n%(file)s\n"
                    "because %(error)s -- recreating it\n"
                ) % {"file": filename, "error": str(err)}
                logging.warning(msg1)
                return

            if oldstyle:
                loader = self._load_oldstyle_section
            else:
                loader = self._load_section

            for section in parser.sections():
                name = section.lower()
                if name not in self.data:
                    self.data[name] = {}
                loader(section, parser)

            clean_up(parser)

    def _load_section(self, section, parser):
        """
        Load a section of an .ini file into self.data
        """
        name = section.lower()
        for option in parser.options(section):
            raw_value = get_raw_value(parser, section, option)
            setting = option.lower()

            try:
                value = safe_eval(raw_value)
            except:
                # most likely exception is SyntaxError but
                # others are possible  ex: '0L' from Python2 days
                value = None

            self._load_setting(name, setting, value)

    def _load_oldstyle_section(self, section, parser):
        """
        Load a section of an .ini file into self.data
        """
        name = section.lower()
        for option in parser.options(section):
            raw_value = get_raw_value(parser, section, option)
            setting = option.lower()

            ####################### Upgrade from oldstyle < 3.2
            # Oldstyle didn't mark setting type, but had it
            # set in preferences. New style gets it from evaling
            # the setting's value
            #######################
            # if we know this setting, convert type
            key = f"{name}.{setting}"
            if self.has_default(key):
                vtype = type(self.get_default(key))
                if vtype == bool:
                    value = raw_value in ["1", "True"]
                elif vtype == list:
                    logging.warning("WARNING: ignoring old key '%s'", key)
                    continue  # there were no lists in oldstyle
                else:
                    value = vtype(raw_value)
            else:
                # else, ignore it
                logging.warning("WARNING: ignoring old key '%s'", key)
                continue  # with next setting
            ####################### End upgrade code
            self._load_setting(name, setting, value)

    def _load_setting(self, name, setting, value):
        if name in self.default and setting in self.default[name]:
            if isinstance(self.default[name][setting], bool):
                value = bool(value)
            if self.check_type(self.default[name][setting], value):
                self.data[name][setting] = value
            else:
                logging.warning(
                    "WARNING: ignoring key with wrong type "
                    "'%s.%s' %s needed instead of %s",
                    name,
                    setting,
                    type(self.data[name][setting]),
                    type(value),
                )
        else:
            # this could be a third-party setting; add it:
            self.data[name][setting] = value

    def save(self, filename=None, comments=None):
        """
        Saves the current section/settings to an .ini file. Optional filename
        will override the default filename to save to, if given.
        """
        if filename is None:
            filename = self.filename
        if filename:
            try:
                head = os.path.split(filename)[0]
                os.makedirs(head)
            except OSError as exp:
                if exp.errno != errno.EEXIST:
                    raise
            try:
                with open(filename, "w", encoding="utf-8") as key_file:
                    key_file.write(";; Gramps key file\n")
                    key_file.write(
                        ";; Automatically created at "
                        f"{time.strftime('%Y/%m/%d %H:%M:%S')}\n\n"
                    )
                    write_comments(key_file, comments)
                    for section in sorted(self.data):
                        key_file.write(f"[{section}]\n")
                        for key in sorted(self.data[section]):
                            value = self.data[section][key]
                            default = ""  # might be a third-party setting
                            if self.has_default(f"{section}.{key}"):
                                if value == self.get_default(f"{section}.{key}"):
                                    default = ";;"
                            if isinstance(value, int):
                                value = int(value)  # TODO why is this needed?
                            key_file.write(f"{default}{key}={repr(value)}\n")
                        key_file.write("\n")
                # else, no filename given; nothing to save so do nothing quietly
            except IOError as err:
                logging.warning("Failed to open %s because %s", filename, str(err))
                return

    def get(self, key):
        """
        Get the setting's value. raise an error if an invalid section.setting.
        Key is a string in the "section.setting" format.
        """
        try:
            section, setting = key.split(".", 1)
        except ValueError as error:
            raise AttributeError(
                f"Invalid config section.setting name: '{key}'"
            ) from error

        self._validate_section_and_setting(section, setting)
        return self.data[section][setting]

    def _validate_section_and_setting(self, section, setting):
        """
        Validate section and setting present in the loaded data.
        """
        if section not in self.data:
            raise AttributeError(f"No such config section name: '{section}'")
        if setting not in self.data[section]:
            raise AttributeError(f"No such config setting name: '{section}.{setting}'")

    def is_set(self, key):
        """
        Does the setting exist? Returns True if does, False otherwise.
        Key is a string in the "section.setting" format.
        """
        try:
            section, setting = key.split(".", 1)
        except ValueError:
            return False

        if section not in self.data:
            return False
        if setting not in self.data[section]:
            return False
        return True

    def has_default(self, key):
        """
        Does the setting have a default value? Returns True if it does,
        False otherwise. Key is a string in the "section.setting" format.
        """
        try:
            section, setting = key.split(".", 1)
        except ValueError:
            return False

        if section not in self.default:
            return False
        if setting not in self.default[section]:
            return False
        return True

    def get_default(self, key):
        """
        Get the setting's default value. Raises an error if invalid key is
        give. Key is a string in the "section.setting" format.
        """
        try:
            section, setting = key.split(".", 1)
        except ValueError as error:
            raise AttributeError(
                f"Invalid config section.setting name: '{key}'"
            ) from error

        if section not in self.default:
            raise AttributeError(f"No such config section name: '{section}'")
        if setting not in self.default[section]:
            raise AttributeError(f"No such config setting name: '{section}.{setting}'")
        return self.default[section][setting]

    def register(self, key, default):
        """
        Register a section.setting, and set the default.
        Will overwrite any previously set default, and set setting if not one.
        The default value deterimines the type of the setting.
        """
        try:
            section, setting = key.split(".", 1)
        except ValueError as error:
            raise AttributeError(
                f"Invalid config section.setting name: '{key}'"
            ) from error

        if section not in self.data:
            self.data[section] = {}
        if section not in self.default:
            self.default[section] = {}
        if section not in self.callbacks:
            self.callbacks[section] = {}
        if setting not in self.callbacks[section]:
            self.callbacks[section][setting] = []
        # Add the default value to settings, if not exist:
        if setting not in self.data[section]:
            self.data[section][setting] = default
        # Set the default, regardless:
        self.default[section][setting] = copy.deepcopy(default)

    def connect(self, key, func):
        """
        Connect a callback func that gets called when key is changed.
        """
        try:
            section, setting = key.split(".", 1)
        except ValueError as error:
            raise AttributeError(
                f"Invalid config section.setting name: '{key}'"
            ) from error

        self._validate_section_and_setting(section, setting)
        self._cb_id += 1
        self.callbacks[section][setting].append((self._cb_id, func))
        return self._cb_id

    def disconnect(self, callback_id):
        """
        Removes a callback given its callback ID. The ID is generated and
        returned when the function is connected to the key (section.setting).
        """
        for dummy_section, settings in self.callbacks.items():
            for dummy_setting, callbacks in settings.items():
                for cbid, func in callbacks:
                    if callback_id == cbid:
                        callbacks.remove((cbid, func))

    def emit(self, key):
        """
        Emits the signal "key" which will call the callbacks associated
        with that setting.
        """
        try:
            section, setting = key.split(".", 1)
        except ValueError as error:
            raise AttributeError(
                f"Invalid config section.setting name: '{key}'"
            ) from error

        if section not in self.callbacks:
            raise AttributeError(f"No such config section name: '{section}'")
        if setting not in self.callbacks[section]:
            raise AttributeError(f"No such config setting name: '{section}.{setting}'")
        for dummy_cbid, func in self.callbacks[section][setting]:
            func(self, 0, str(self.data[section][setting]), None)

    def set(self, key, value):
        """
        Set the setting's value. There are only two ways to get into
        the data dictionary: via the :meth:`load` method that reads a file,
        or from this method.
        """
        try:
            section, setting = key.split(".", 1)
        except ValueError as error:
            raise AttributeError(
                "Invalid config section.setting name: '{key}'"
            ) from error

        self._validate_section_and_setting(section, setting)

        # Check value to see if right type:
        if self.has_default(key):
            if not self.check_type(self.get_default(key), value):
                raise AttributeError(
                    f"attempting to set '{key}' to wrong type "
                    f"'{type(value)}'; should be "
                    f"'{type(self.get_default(key))}'"
                )
        if setting in self.data[section] and self.data[section][setting] == value:
            # Do nothing if existed and is the same
            pass
        else:
            # Set the value:
            self.data[section][setting] = value
            # Only call callback if the value changed!
            if section in self.callbacks and setting in self.callbacks[section]:
                self.emit(key)

    def check_type(self, value1, value2):
        """
        Check if value1 and value2 are different types.
        """
        type1 = type(value1)
        type2 = type(value2)
        if type1 == type2:
            return True
        if isinstance(value1, str) and isinstance(value2, str):
            return True
        if type1 in [int, float] and type2 in [int, float]:
            return True
        return False


def get_raw_value(parser, section, option):
    """
    Prepare and return raw value.
    """
    raw_value = parser.get(section, option).strip()
    if raw_value[:2] == "u'":
        raw_value = raw_value[1:]
    elif raw_value.startswith("["):
        raw_value = raw_value.replace(", u'", ", '")
        raw_value = raw_value.replace("[u'", "['")
    return raw_value


def write_comments(output_file, comments):
    """
    Sanity check and write comments out to a .ini file.
    """
    if comments:
        if not isinstance(comments, list):
            raise AttributeError("Comments should be a list")

        output_file.write("\n")
        for comment in comments:
            if not isinstance(comment, str):
                raise AttributeError("Comment should be a string")
            clean_comment = comment.strip("; \n")
            output_file.write(f";; {clean_comment}\n")



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2009       Brian G. Matherly
# Copyright (C) 2009       Gary Burton
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Provide the management of databases. This includes opening, renaming,
creating, and deleting of databases.
"""

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import os
import time
import copy
import subprocess
from urllib.parse import urlparse
import logging
import re

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import Pango

# -------------------------------------------------------------------------
#
# gramps modules
#
# -------------------------------------------------------------------------
from .display import display_help
from gramps.gen.const import URL_WIKISTRING, URL_MANUAL_PAGE
from .user import User
from .dialog import ErrorDialog, QuestionDialog, QuestionDialog2, ICON
from .pluginmanager import GuiPluginManager
from gramps.cli.clidbman import CLIDbManager, NAME_FILE, time_val, UNAVAILABLE
from .managedwindow import ManagedWindow
from .ddtargets import DdTargets
from gramps.gen.recentfiles import rename_filename, remove_filename
from .glade import Glade
from gramps.gen.db.exceptions import DbException
from gramps.gen.db.utils import make_database, open_database
from gramps.gen.config import config
from .listmodel import ListModel
from gramps.gen.constfunc import win
from gramps.gen.plug import BasePluginManager
from gramps.gen.const import GRAMPS_LOCALE as glocale
from gramps.gui.widgets.persistenttreeview import PersistentTreeView

_ = glocale.translation.gettext

# -------------------------------------------------------------------------
#
# set up logging
#
# -------------------------------------------------------------------------
LOG = logging.getLogger(".DbManager")

# -------------------------------------------------------------------------
#
# constants
#
# -------------------------------------------------------------------------
if win():
    _RCS_FOUND = os.system("rcs -V >nul 2>nul") == 0
    if _RCS_FOUND and "TZ" not in os.environ:
        # RCS requires the "TZ" variable be set.
        os.environ["TZ"] = str(time.timezone)
else:
    _RCS_FOUND = os.system("rcs -V >/dev/null 2>/dev/null") == 0

_RETURN = Gdk.keyval_from_name("Return")
_KP_ENTER = Gdk.keyval_from_name("KP_Enter")

WIKI_HELP_PAGE = _("%s_-_Manage_Family_Trees") % URL_MANUAL_PAGE
WIKI_HELP_SEC = _("Family_Trees_manager_window")

ARCHIVE = "rev.gramps"
ARCHIVE_V = "rev.gramps,v"

NAME_COL = 0
PATH_COL = 1
FILE_COL = 2
DATE_COL = 3
DSORT_COL = 4
OPEN_COL = 5
ICON_COL = 6
BACKEND_COL = 7

RCS_BUTTON = {True: _("_Extract"), False: _("_Archive")}


class Information(ManagedWindow):
    def __init__(self, uistate, data, track):
        super().__init__(uistate, track, self, modal=True)
        self.window = Gtk.Dialog()
        self.set_window(self.window, None, _("Database Information"))
        self.setup_configs("interface.information", 600, 400)
        self.ok = self.window.add_button(_("_OK"), Gtk.ResponseType.OK)
        self.ok.connect("clicked", self.on_ok_clicked)
        s = Gtk.ScrolledWindow()
        titles = [(_("Setting"), 0, 150), (_("Value"), 1, 400)]
        treeview = Gtk.TreeView()
        model = ListModel(treeview, titles)
        for key, value in sorted(data.items()):
            model.add(
                (
                    key,
                    str(value),
                ),
                key,
            )
        s.add(treeview)
        self.window.vbox.pack_start(s, True, True, 0)
        self.show()

    def on_ok_clicked(self, obj):
        self.window.close()

    def build_menu_names(self, obj):
        return (_("Database Information"), None)


class DbManager(CLIDbManager, ManagedWindow):
    """
    Database Manager. Opens a database manager window that allows users to
    create, rename, delete and open databases.
    """

    ICON_MAP = {
        CLIDbManager.ICON_NONE: None,
        CLIDbManager.ICON_RECOVERY: "dialog-error",
        CLIDbManager.ICON_LOCK: "gramps-lock",
        CLIDbManager.ICON_OPEN: "document-open",
    }

    BUSY_CURSOR = Gdk.Cursor.new_for_display(
        Gdk.Display.get_default(), Gdk.CursorType.WATCH
    )

    def __init__(self, uistate, dbstate, viewmanager, parent=None):
        """
        Create the top level window from the glade description, and extracts
        the GTK widgets that are needed.
        """
        window_id = self
        ManagedWindow.__init__(self, uistate, [], window_id, modal=True)
        CLIDbManager.__init__(self, dbstate)
        self.glade = Glade(toplevel="dbmanager")
        self.top = self.glade.toplevel
        self.set_window(self.top, None, None)
        self.setup_configs("interface.dbmanager", 780, 350)
        self.viewmanager = viewmanager

        for attr in [
            "connect_btn",
            "cancel_btn",
            "new_btn",
            "remove_btn",
            "info_btn",
            "rename_btn",
            "convert_btn",
            "repair_btn",
            "rcs_btn",
            "msg",
            "close_btn",
        ]:
            setattr(self, attr, self.glade.get_object(attr))

        self.model = None
        self.column = None
        self.lock_file = None
        self.data_to_delete = None

        self.dblist = PersistentTreeView(uistate, "dbman")
        self.dblist.set_vexpand(True)
        scrolledwindow = self.glade.get_object("scrolledwindow")
        scrolledwindow.add(self.dblist)
        self.selection = self.dblist.get_selection()

        # For already loaded database:
        self._current_node = None
        self.__connect_signals()
        self.__build_interface()
        self._populate_model()
        self.before_change = ""
        self.after_change = ""
        self._select_default()
        self.user = User(
            error=ErrorDialog,
            parent=parent,
            callback=self.uistate.pulse_progressbar,
            uistate=self.uistate,
        )

    def build_menu_names(self, obj):
        """This window can have children, but they are modal so no submenu
        is visible"""
        submenu_label = " "
        menu_label = _("Family Trees")
        return (menu_label, submenu_label)

    def _select_default(self):
        """
        Select the current, or latest, tree.
        """
        # If already loaded database, center on it:
        if self._current_node:
            store, node = self.selection.get_selected()
            tree_path = store.get_path(self._current_node)
            self.selection.select_path(tree_path)
            self.dblist.scroll_to_cell(tree_path, None, 1, 0.5, 0)

    def __connect_signals(self):
        """
        Connects the signals to the buttons on the interface.
        """
        ddtarget = DdTargets.URI_LIST
        self.top.drag_dest_set(
            Gtk.DestDefaults.ALL, [DdTargets.URI_LIST.target()], Gdk.DragAction.COPY
        )

        self.remove_btn.connect("clicked", self.__remove_db)
        self.new_btn.connect("clicked", self.__new_db)
        self.rename_btn.connect("clicked", self.__rename_db)
        self.convert_btn.connect("clicked", self.__convert_db_ask)
        self.info_btn.connect("clicked", self.__info_db)
        self.close_btn.connect("clicked", self.__close_db)
        self.repair_btn.connect("clicked", self.__repair_db)
        self.selection.connect("changed", self.__selection_changed)
        self.dblist.connect("button-press-event", self.__button_press)
        self.dblist.connect("key-press-event", self.__key_press)
        self.top.connect("drag_data_received", self.__drag_data_received)
        self.top.connect("drag_motion", drag_motion)
        self.top.connect("drag_drop", drop_cb)
        self.define_help_button(
            self.glade.get_object("help_btn"), WIKI_HELP_PAGE, WIKI_HELP_SEC
        )

        if _RCS_FOUND:
            self.rcs_btn.connect("clicked", self.__rcs)

    def define_help_button(self, button, webpage="", section=""):
        button.connect("clicked", lambda x: display_help(webpage, section))

    def __button_press(self, obj, event):
        """
        Checks for a double click event. In the tree view, we want to
        treat a double click as if it was OK button press. However, we have
        to make sure that an item was selected first.
        """
        if event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS and event.button == 1:
            if self.connect_btn.get_property("sensitive"):
                self.top.response(Gtk.ResponseType.OK)
                return True
        return False

    def __key_press(self, obj, event):
        """
        Grab ENTER so it does not start editing the cell, but behaves
        like double click instead
        """
        if event.keyval in (_RETURN, _KP_ENTER):
            if self.connect_btn.get_property("sensitive"):
                self.top.response(Gtk.ResponseType.OK)
                return True
        return False

    def __selection_changed(self, selection):
        """
        Called when the selection is changed in the TreeView.
        """
        self.__update_buttons(selection)

    def __update_buttons(self, selection):
        """
        What we are trying to detect is the selection or unselection of a row.
        When a row is unselected, the Open, Rename, and Remove buttons
        are set insensitive. If a row is selected, the rename and remove
        buttons are disabled, and the Open button is disabled if the
        row represents a open database.
        """

        # Get the current selection
        store, node = selection.get_selected()

        if not _RCS_FOUND:  # it's not in Windows
            self.rcs_btn.set_visible(False)

        # if nothing is selected
        if not node:
            self.connect_btn.set_sensitive(False)
            self.rename_btn.set_sensitive(False)
            self.convert_btn.set_sensitive(False)
            self.info_btn.set_sensitive(False)
            self.close_btn.set_sensitive(False)
            self.rcs_btn.set_sensitive(False)
            self.repair_btn.set_sensitive(False)
            self.remove_btn.set_sensitive(False)
            return

        path = self.model.get_path(node)
        if path is None:
            return

        is_rev = len(path.get_indices()) > 1
        self.rcs_btn.set_label(RCS_BUTTON[is_rev])

        if store.get_value(node, ICON_COL) == "document-open":
            self.close_btn.set_sensitive(True)
            self.convert_btn.set_sensitive(False)
            self.connect_btn.set_sensitive(False)
            if _RCS_FOUND:
                self.rcs_btn.set_sensitive(True)
        elif store.get_value(node, BACKEND_COL) == UNAVAILABLE:
            self.close_btn.set_sensitive(False)
            self.convert_btn.set_sensitive(False)
            self.connect_btn.set_sensitive(False)
            self.rcs_btn.set_sensitive(False)
            self.repair_btn.set_sensitive(False)
        else:
            self.close_btn.set_sensitive(False)
            dbid = config.get("database.backend")
            backend_type = self.get_backend_name_from_dbid(dbid)
            if (
                store.get_value(node, ICON_COL) in [None, ""]
                and store.get_value(node, BACKEND_COL) != backend_type
            ):
                self.convert_btn.set_sensitive(True)
            else:
                self.convert_btn.set_sensitive(False)
            self.connect_btn.set_sensitive(not is_rev)
            if _RCS_FOUND and is_rev:
                self.rcs_btn.set_sensitive(True)
            else:
                self.rcs_btn.set_sensitive(False)

        if store.get_value(node, ICON_COL) == "dialog-error":
            path = store.get_value(node, PATH_COL)
            backup = os.path.join(path, "person.gbkp")
            self.repair_btn.set_sensitive(os.path.isfile(backup))
        else:
            self.repair_btn.set_sensitive(False)

        self.rename_btn.set_sensitive(True)
        self.info_btn.set_sensitive(True)
        self.remove_btn.set_sensitive(True)
        self.new_btn.set_sensitive(True)

    def __build_interface(self):
        """
        Builds the columns for the TreeView. The columns are:

        Icon, Database Name, Last Modified, Backend Type

        The Icon column gets its data from column 6 of the database model.
        It is expecting either None, or a GTK stock icon name

        The Database Name column is an editable column. We connect to the
        'edited' signal, so that we can change the name when the user changes
        the column.

        The last accessed column simply displays the last time famtree was
        opened.

        The Backend Type column is a string based on database backend.
        """
        # Put some help on the buttons:
        dbid = config.get("database.backend")
        backend_type = self.get_backend_name_from_dbid(dbid)
        if backend_type == UNAVAILABLE:
            dbid = "sqlite"
            config.set("database.backend", dbid)
            backend_type = self.get_backend_name_from_dbid(dbid)
        self.new_btn.set_tooltip_text(backend_type)

        # build the database name column
        render = Gtk.CellRendererText()
        render.set_property("ellipsize", Pango.EllipsizeMode.END)
        render.connect("edited", self.__change_name)
        render.connect("editing-canceled", self.__stop_edit)
        render.connect("editing-started", self.__start_edit)
        self.column = Gtk.TreeViewColumn(_("Family Tree name"), render, text=NAME_COL)
        self.column.set_sort_column_id(NAME_COL)
        self.column.set_sort_indicator(True)
        self.column.set_resizable(True)
        self.column.set_min_width(250)
        self.dblist.append_column(self.column)
        self.name_renderer = render

        # build the icon column
        render = Gtk.CellRendererPixbuf()
        # icon_column = Gtk.TreeViewColumn(_('Status'), render,
        # icon_name=ICON_COL)
        icon_column = Gtk.TreeViewColumn(_("Status"), render)
        icon_column.set_cell_data_func(render, bug_fix)
        icon_column.set_sort_column_id(ICON_COL)
        self.dblist.append_column(icon_column)

        # build the backend column
        render = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn(_("Database Type"), render, text=BACKEND_COL)
        column.set_sort_column_id(BACKEND_COL)
        column.set_sort_indicator(True)
        column.set_resizable(True)
        self.dblist.append_column(column)

        # build the last accessed column
        render = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn(_("Last accessed"), render, text=DATE_COL)
        column.set_sort_column_id(DSORT_COL)
        self.dblist.append_column(column)
        self.dblist.restore_column_size()

    def __populate(self):
        """
        Builds the data and the display model.
        """
        self._populate_cli()
        self._populate_model()

    def _populate_model(self):
        """
        Builds the display model.
        """
        self.model = Gtk.TreeStore(str, str, str, str, int, bool, str, str)

        # use current names to set up the model
        self._current_node = None
        last_accessed_node = None
        last_accessed = 0
        for items in self.current_names:
            data = list(items[:8])
            backend_type = self.get_backend_name_from_dbid(data[BACKEND_COL])
            node = self.model.append(None, data[:-1] + [backend_type])
            # For already loaded database, set current_node:
            if self.dbstate.is_open() and self.dbstate.db.get_save_path() == data[1]:
                self._current_node = node
            if data[DSORT_COL] > last_accessed:
                last_accessed = data[DSORT_COL]
                last_accessed_node = node
            for rdata in find_revisions(os.path.join(items[1], ARCHIVE_V)):
                data = [
                    rdata[2],
                    rdata[0],
                    items[1],
                    rdata[1],
                    0,
                    False,
                    "",
                    backend_type,
                ]
                self.model.append(node, data)
        if self._current_node is None:
            self._current_node = last_accessed_node
        self.model.set_sort_column_id(NAME_COL, Gtk.SortType.ASCENDING)
        self.dblist.set_model(self.model)

    def existing_name(self, name, skippath=None):
        """
        Return true if a name is present in the model already.
        If skippath given, the name of skippath is not considered
        """
        iter = self.model.get_iter_first()
        while iter:
            path = self.model.get_path(iter)
            if path == skippath:
                pass
            else:
                itername = self.model.get_value(iter, NAME_COL)
                if itername.strip() == name.strip():
                    return True
            iter = self.model.iter_next(iter)
        return False

    def run(self):
        """
        Runs the dialog, returning None if nothing has been chosen,
        or the path and name if something has been selected
        """
        self.show()
        self.__update_buttons(self.selection)
        while True:
            value = self.top.run()
            if value == Gtk.ResponseType.OK:
                store, node = self.selection.get_selected()
                # don't open a locked file
                if store.get_value(node, ICON_COL) == "gramps-lock":
                    self.__ask_to_break_lock(store, node)
                    continue
                # don't open a version
                if len(store.get_path(node).get_indices()) > 1:
                    continue
                if node:
                    del self.selection
                    del self.name_renderer
                    self.close()
                    path = store.get_value(node, PATH_COL)
                    return (path, store.get_value(node, NAME_COL))
            else:
                del self.selection
                del self.name_renderer
                if value != Gtk.ResponseType.DELETE_EVENT:
                    self.close()
                return None

    def __ask_to_break_lock(self, store, node):
        """
        Prompts the user for permission to break the lock file that another
        process has set on the file.
        """
        path = store.get_path(node)
        self.lock_file = store[path][PATH_COL]

        QuestionDialog(
            _("Break the lock on the '%s' database?") % store[path][0],
            _(
                "Gramps believes that someone else is actively editing "
                "this database. You cannot edit this database while it "
                "is locked. If no one is editing the database you may "
                "safely break the lock. However, if someone else is editing "
                "the database and you break the lock, you may corrupt the "
                "database."
            ),
            _("Break lock"),
            self.__really_break_lock,
            parent=self.top,
        )

    def __really_break_lock(self):
        """
        Deletes the lock file associated with the selected database,
        then updates the display appropriately.
        """
        try:
            self.break_lock(self.lock_file)
            store, node = self.selection.get_selected()
            dbpath = store.get_value(node, PATH_COL)
            (tval, last) = time_val(dbpath)
            store.set_value(node, OPEN_COL, 0)
            store.set_value(node, ICON_COL, "")  # see bug_fix
            store.set_value(node, DATE_COL, last)
            store.set_value(node, DSORT_COL, tval)
        except IOError:
            return

    def __stop_edit(self, *args):
        self.name_renderer.set_property("editable", False)
        self.__update_buttons(self.selection)

    def __start_edit(self, *args):
        """
        Do not allow to click Load while changing name, to force users to finish
        the action of renaming. Hack around the fact that clicking button
        sends a 'editing-canceled' signal loosing the new name
        """
        self.connect_btn.set_sensitive(False)
        self.rename_btn.set_sensitive(False)
        self.convert_btn.set_sensitive(False)
        self.info_btn.set_sensitive(False)
        self.rcs_btn.set_sensitive(False)
        self.repair_btn.set_sensitive(False)
        self.remove_btn.set_sensitive(False)
        self.new_btn.set_sensitive(False)

    def __change_name(self, renderer_sel, path, new_text):
        """
        Change the name of the database. This is a callback from the
        column, which has been marked as editable.

        If the new string is empty, do nothing. Otherwise, renaming the
        database is simply changing the contents of the name file.
        """
        # kill special characters so can use as file name in backup.
        new_text = re.sub(r"[':<>|,;=\"\[\]\.\+\*\/\?\\]", "_", new_text)
        # path is a string, convert to TreePath first
        path = Gtk.TreePath(path=path)
        if len(new_text) > 0:
            node = self.model.get_iter(path)
            old_text = self.model.get_value(node, NAME_COL)
            if self.model.get_value(node, ICON_COL) == "document-open":
                # this database is loaded. We must change the title
                # in case we change the name several times before quitting,
                # we save the first old name.
                if self.before_change == "":
                    self.before_change = old_text
                self.after_change = new_text
            if not old_text.strip() == new_text.strip():
                if len(path.get_indices()) > 1:
                    self.__rename_revision(path, new_text)
                else:
                    self.__rename_database(path, new_text)

        self.name_renderer.set_property("editable", False)
        self.__update_buttons(self.selection)

    def __rename_revision(self, path, new_text):
        """
        Renames the RCS revision using the rcs command. The rcs command
        is in the format of:

           rcs -mREV:NEW_NAME archive

        """
        node = self.model.get_iter(path)
        db_dir = self.model.get_value(node, FILE_COL)
        rev = self.model.get_value(node, PATH_COL)
        archive = os.path.join(db_dir, ARCHIVE_V)

        cmd = ["rcs", "-x,v", "-m%s:%s" % (rev, new_text), archive]

        proc = subprocess.Popen(cmd, stderr=subprocess.PIPE)
        status = proc.wait()
        message = "\n".join(proc.stderr.readlines())
        proc.stderr.close()
        del proc

        if status != 0:
            ErrorDialog(
                _("Rename failed"),
                _(
                    "An attempt to rename a version failed "
                    "with the following message:\n\n%s"
                )
                % message,
                parent=self.top,
            )
        else:
            self.model.set_value(node, NAME_COL, new_text)
            # scroll to new position
            store, node = self.selection.get_selected()
            tree_path = store.get_path(node)
            self.dblist.scroll_to_cell(tree_path, None, False, 0.5, 0.5)

    def __rename_database(self, path, new_text):
        """
        Renames the database by writing the new value to the name.txt file
        """
        new_text = new_text.strip()
        node = self.model.get_iter(path)
        filename = self.model.get_value(node, FILE_COL)
        if self.existing_name(new_text, skippath=path):
            ErrorDialog(
                _("Could not rename the Family Tree."),
                _("Family Tree already exists, choose a unique name."),
                parent=self.top,
            )
            return
        old_text, new_text = self.rename_database(filename, new_text)
        if old_text is not None:
            rename_filename(old_text, new_text)
            self.model.set_value(node, NAME_COL, new_text)
        # scroll to new position
        store, node = self.selection.get_selected()
        tree_path = store.get_path(node)
        self.dblist.scroll_to_cell(tree_path, None, False, 0.5, 0.5)

    def __rcs(self, obj):
        """
        Callback for the RCS button. If the tree path is > 1, then we are
        on an RCS revision, in which case we can check out. If not, then
        we can only check in.
        """
        store, node = self.selection.get_selected()
        tree_path = store.get_path(node)
        if len(tree_path.get_indices()) > 1:
            parent_node = store.get_iter((tree_path[0],))
            parent_name = store.get_value(parent_node, NAME_COL)
            name = store.get_value(node, NAME_COL)
            revision = store.get_value(node, PATH_COL)
            db_path = store.get_value(node, FILE_COL)

            self.__checkout_copy(parent_name, name, revision, db_path)
        else:
            base_path = self.dbstate.db.get_save_path()
            archive = os.path.join(base_path, ARCHIVE)
            _check_in(
                self.dbstate.db,
                archive,
                self.user,
                self.__start_cursor,
                parent=self.window,
            )
            self.__end_cursor()

        self.__populate()
        self._select_default()

    def __checkout_copy(self, parent_name, name, revision, db_path):
        """
        Create a new database, then extracts a revision from RCS and
        imports it into the db
        """
        dbid = config.get("database.backend")
        new_path, newname = self._create_new_db(
            "%s : %s" % (parent_name, name), dbid=dbid
        )

        self.__start_cursor(_("Extracting archive..."))

        dbase = make_database(dbid)
        dbase.load(new_path)

        self.__start_cursor(_("Importing archive..."))
        check_out(dbase, revision, db_path, self.user)
        self.__end_cursor()
        dbase.close(user=self.user)

    def __remove_db(self, obj):
        """
        Callback associated with the Remove button. Get the selected
        row and data, then call the verification dialog.
        """
        store, node = self.selection.get_selected()
        path = store.get_path(node)
        self.data_to_delete = store[path]

        if len(path.get_indices()) == 1:
            QuestionDialog(
                _("Remove the '%s' Family Tree?") % self.data_to_delete[0],
                _("Removing this Family Tree will permanently destroy " "the data."),
                _("Remove Family Tree"),
                self.__really_delete_db,
                parent=self.top,
            )
        else:
            rev = self.data_to_delete[0]
            parent = store[(path[0],)][0]
            QuestionDialog(
                _("Remove the '%(revision)s' version " "of '%(database)s'")
                % {"revision": rev, "database": parent},
                _(
                    "Removing this version will prevent you from "
                    "extracting it in the future."
                ),
                _("Remove version"),
                self.__really_delete_version,
                parent=self.top,
            )

    def __really_delete_db(self):
        """
        Delete the selected database. If the database is open, close it first.
        Then scan the database directory, deleting the files, and finally
        removing the directory.
        """

        # close the database if the user has requested to delete the
        # active database
        if self.data_to_delete[PATH_COL] == self.active:
            self.uistate.viewmanager.close_database()

        store, node = self.selection.get_selected()
        path = store.get_path(node)
        node = self.model.get_iter(path)
        filename = self.model.get_value(node, FILE_COL)
        try:
            with open(filename, "r", encoding="utf-8") as name_file:
                file_name_to_delete = name_file.read()
            remove_filename(file_name_to_delete)
            directory = self.data_to_delete[1]
            for top, dirs, files in os.walk(directory):
                for filename in files:
                    os.unlink(os.path.join(top, filename))
            os.rmdir(directory)
        except (IOError, OSError) as msg:
            ErrorDialog(_("Could not delete Family Tree"), str(msg), parent=self.top)
        # rebuild the display
        self.__populate()
        self._select_default()

    def __really_delete_version(self):
        """
        Delete the selected database. If the database is open, close it first.
        Then scan the database directory, deleting the files, and finally
        removing the directory.
        """
        db_dir = self.data_to_delete[FILE_COL]
        rev = self.data_to_delete[PATH_COL]
        archive = os.path.join(db_dir, ARCHIVE_V)

        cmd = ["rcs", "-x,v", "-o%s" % rev, "-q", archive]

        proc = subprocess.Popen(cmd, stderr=subprocess.PIPE)
        status = proc.wait()
        message = "\n".join(proc.stderr.readlines())
        proc.stderr.close()
        del proc

        if status != 0:
            ErrorDialog(
                _("Deletion failed"),
                _(
                    "An attempt to delete a version failed "
                    "with the following message:\n\n%s"
                )
                % message,
                parent=self.top,
            )

        # rebuild the display
        self.__populate()
        self._select_default()

    def __convert_db_ask(self, obj):
        """
        Ask to convert a closed family tree into the default database backend.
        """
        store, node = self.selection.get_selected()
        name = store[node][0]
        dirname = store[node][1]
        dbid = config.get("database.backend")
        backend_type = self.get_backend_name_from_dbid(dbid)
        QuestionDialog(
            _("Convert the '%s' database?") % name,
            _(
                "Do you wish to convert this family tree into a "
                "%(database_type)s database?"
            )
            % {"database_type": backend_type},
            _("Convert"),
            lambda: self.__convert_db(name, dirname),
            parent=self.top,
        )

    def __convert_db(self, name, dirname):
        """
        Actually convert the family tree into the default database backend.
        """
        try:
            db = open_database(name)
        except:
            ErrorDialog(
                _("Opening the '%s' database") % name,
                _(
                    "An attempt to convert the database failed. "
                    "Perhaps it needs updating."
                ),
                parent=self.top,
            )
            return
        plugin_manager = GuiPluginManager.get_instance()
        export_function = None
        for plugin in plugin_manager.get_export_plugins():
            if plugin.get_extension() == "gramps":
                export_function = plugin.get_export_function()
                break
        ## Next, get an XML dump:
        if export_function is None:
            ErrorDialog(
                _("Converting the '%s' database") % name,
                _("An attempt to export the database failed."),
                parent=self.top,
            )
            db.close(user=self.user)
            return
        self.__start_cursor(_("Converting data..."))
        xml_file = os.path.join(dirname, "backup.gramps")
        export_function(db, xml_file, self.user)
        db.close(user=self.user)
        count = 1
        new_text = "%s %s" % (name, _("(Converted #%d)") % count)
        while self.existing_name(new_text):
            count += 1
            new_text = "%s %s" % (name, _("(Converted #%d)") % count)
        dbid = config.get("database.backend")
        new_path, newname = self._create_new_db(new_text, dbid=dbid, edit_entry=False)
        ## Create a new database of correct type:
        dbase = make_database(dbid)
        dbase.load(new_path)
        ## import from XML
        import_function = None
        for plugin in plugin_manager.get_import_plugins():
            if plugin.get_extension() == "gramps":
                import_function = plugin.get_import_function()
        if import_function is None:
            ErrorDialog(
                _("Converting the '%s' database") % name,
                _("An attempt to import into the database failed."),
                parent=self.top,
            )
        else:
            import_function(dbase, xml_file, self.user)
        self.__end_cursor()
        dbase.close(user=self.user)
        self.__populate()
        self._select_default()

    def __rename_db(self, obj):
        """
        Start the rename process by calling the start_editing option on
        the line with the cursor.
        """
        store, node = self.selection.get_selected()
        path = self.model.get_path(node)
        self.name_renderer.set_property("editable", True)
        self.dblist.set_cursor(path, self.column, True)

    def __close_db(self, obj):
        """
        Close the database. Set the displayed line correctly, set the dbstate to
        no_database, update the sensitivity of the buttons in this dialogue box
        and get viewmanager to manage the main window and plugable views.
        """
        store, node = self.selection.get_selected()
        dbpath = store.get_value(node, PATH_COL)
        (tval, last) = time_val(dbpath)
        store.set_value(node, OPEN_COL, 0)
        store.set_value(node, ICON_COL, "")  # see bug_fix
        store.set_value(node, DATE_COL, last)
        store.set_value(node, DSORT_COL, tval)
        self.dbstate.no_database()
        self.__update_buttons(self.selection)
        self.viewmanager.post_close_db()

    def __info_db(self, obj):
        """
        Show info on this database.
        """
        store, node = self.selection.get_selected()
        name = store[node][0]
        dirname = store[node][1]
        # if this is open, get info from there, otherwise, temp open?
        summary = self.get_dbdir_summary(dirname, name)
        Information(self.uistate, summary, track=self.track)

    def __repair_db(self, obj):
        """
        Start the repair process by calling the start_editing option on
        the line with the cursor.
        """
        store, node = self.selection.get_selected()
        dirname = store[node][1]

        # First ask user if he is really sure :-)
        yes_no = QuestionDialog2(
            _("Repair Family Tree?"),
            _(
                "If you click %(bold_start)sProceed%(bold_end)s, Gramps will "
                "attempt to recover your Family Tree from the last good "
                "backup. There are several ways this can cause unwanted "
                "effects, so %(bold_start)sbackup%(bold_end)s the "
                "Family Tree first.\nThe Family Tree you have selected "
                "is stored in %(dirname)s.\n\n"
                "Before doing a repair, verify that the Family Tree can "
                "really no longer be opened, as the database back-end can "
                "recover from some errors automatically.\n\n"
                "%(bold_start)sDetails:%(bold_end)s Repairing a Family Tree "
                "actually uses the last backup of the Family Tree, which "
                "Gramps stored on last use. If you have worked for "
                "several hours/days without closing Gramps, then all "
                "this information will be lost! If the repair fails, then "
                "the original Family Tree will be lost forever, hence "
                "a backup is needed. If the repair fails, or too much "
                "information is lost, you can fix the original "
                "Family Tree manually. For details, see the webpage\n"
                "%(gramps_wiki_recover_url)s\n"
                "Before doing a repair, try to open the Family Tree "
                "in the normal manner. Several errors that trigger the "
                "repair button can be fixed automatically. "
                "If this is the case, you can disable the repair button "
                "by removing the file %(recover_file)s in the "
                "Family Tree directory."
            )
            % {
                "bold_start": "<b>",
                "bold_end": "</b>",
                "recover_file": "<i>need_recover</i>",
                "gramps_wiki_recover_url": URL_WIKISTRING
                + "Recover_corrupted_family_tree",
                "dirname": dirname,
            },
            _("Proceed, I have taken a backup"),
            _("Stop"),
            parent=self.top,
        )
        prompt = yes_no.run()
        if not prompt:
            return

        opened = store[node][OPEN_COL]
        if opened:
            self.dbstate.no_database()

        # delete files that are not backup files or the .txt file
        for filename in os.listdir(dirname):
            if os.path.splitext(filename)[1] not in (".gbkp", ".txt"):
                fname = os.path.join(dirname, filename)
                os.unlink(fname)

        dbase = make_database("sqlite")
        dbase.load(dirname, None)

        self.__start_cursor(_("Rebuilding database from backup files"))

        try:
            dbase.restore()
        except DbException as msg:
            ErrorDialog(_("Error restoring backup data"), msg, parent=self.top)

        self.__end_cursor()

        dbase.close(user=self.user)
        self.dbstate.no_database()
        self.__populate()
        self._select_default()

    def __start_cursor(self, msg):
        """
        Set the cursor to the busy state, and displays the associated
        message
        """
        self.msg.set_label(msg)
        self.top.get_window().set_cursor(self.BUSY_CURSOR)
        while Gtk.events_pending():
            Gtk.main_iteration()

    def __end_cursor(self):
        """
        Set the cursor back to normal and clears the message
        """
        self.top.get_window().set_cursor(None)
        self.msg.set_label("")

    def __new_db(self, obj):
        """
        Callback wrapper around the actual routine that creates the
        new database. Catch OSError and IOError and display a warning
        message.
        """
        self.new_btn.set_sensitive(False)
        dbid = config.get("database.backend")
        if dbid:
            try:
                self._create_new_db(dbid=dbid)
            except (OSError, IOError) as msg:
                ErrorDialog(
                    _("Could not create Family Tree"), str(msg), parent=self.top
                )
        self.new_btn.set_sensitive(True)

    def _create_new_db(self, title=None, create_db=True, dbid=None, edit_entry=True):
        """
        Create a new database, append to model
        """
        new_path, title = self.create_new_db_cli(title, create_db, dbid)
        path_name = os.path.join(new_path, NAME_FILE)
        (tval, last) = time_val(new_path)
        backend_type = self.get_backend_name_from_dbid(dbid)
        node = self.model.append(
            None, [title, new_path, path_name, last, tval, False, "", backend_type]
        )
        self.selection.select_iter(node)
        path = self.model.get_path(node)
        if edit_entry:
            self.name_renderer.set_property("editable", True)
            self.dblist.set_cursor(path, self.column, True)
        return new_path, title

    def __drag_data_received(self, widget, context, xpos, ypos, selection, info, rtime):
        """
        Handle the reception of drag data
        """
        drag_value = selection.get_data().decode().strip(" \r\n\x00")
        fname = None
        type = None
        title = None
        # Allow any type of URL ("file://", "http://", etc):
        if drag_value and urlparse(drag_value).scheme != "":
            fname, title = [], []
            for treename in [
                v.strip() for v in drag_value.split("\n") if v.strip() != ""
            ]:
                f, t = self.import_new_db(treename, self.user)
                fname.append(f)
                title.append(t)
        return fname, title


def drag_motion(wid, context, xpos, ypos, time_stamp):
    """
    DND callback that is called on a DND drag motion begin
    """
    Gdk.drag_status(context, Gdk.DragAction.COPY, time_stamp)
    return True


def drop_cb(wid, context, xpos, ypos, time_stamp):
    """
    DND callback that finishes the DND operation
    """
    Gtk.drag_finish(context, True, False, time_stamp)
    return True


def find_revisions(name):
    """
    Finds all the revisions of the specified RCS archive.
    """
    import re

    rev = re.compile(r"\s*revision\s+([\d\.]+)")
    date = re.compile(r"date:\s+(\d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\d)[-+]\d\d;")

    if not os.path.isfile(name) or not _RCS_FOUND:
        return []

    rlog = ["rlog", "-x,v", "-zLT", name]

    proc = subprocess.Popen(rlog, stdout=subprocess.PIPE)
    proc.wait()

    revlist = []
    date_str = ""
    rev_str = ""
    com_str = ""

    get_next = False
    if os.path.isfile(name):
        for line in proc.stdout:
            if not isinstance(line, str):
                # we assume utf-8 ...
                line = line.decode("utf-8")
            match = rev.match(line)
            if match:
                rev_str = copy.copy(match.groups()[0])
                continue
            match = date.match(line)
            if match:
                date_str = time.strftime(
                    "%x %X", time.strptime(match.groups()[0], "%Y-%m-%d %H:%M:%S")
                )
                get_next = True
                continue
            if get_next:
                get_next = False
                com_str = line.strip()
                revlist.append((rev_str, date_str, com_str))
    proc.stdout.close()
    del proc
    return revlist


def check_out(dbase, rev, path, user):
    """
    Checks out the revision from rcs, and loads the resulting XML file
    into the database.
    """
    co_cmd = ["co", "-x,v", "-q%s" % rev] + [
        os.path.join(path, ARCHIVE),
        os.path.join(path, ARCHIVE_V),
    ]

    proc = subprocess.Popen(co_cmd, stderr=subprocess.PIPE)
    status = proc.wait()
    message = "\n".join(proc.stderr.readlines())
    proc.stderr.close()
    del proc

    if status != 0:
        user.notify_error(
            _("Retrieve failed"),
            _(
                "An attempt to retrieve the data failed "
                "with the following message:\n\n%s"
            )
            % message,
        )
        return

    pmgr = GuiPluginManager.get_instance()
    for plugin in pmgr.get_import_plugins():
        if plugin.get_extension() == "gramps":
            rdr = plugin.get_import_function()

    xml_file = os.path.join(path, ARCHIVE)
    rdr(dbase, xml_file, user)
    os.unlink(xml_file)


def _check_in(dbase, filename, user, cursor_func=None, parent=None):
    """
    Checks in the specified file into RCS
    """
    init = ["rcs", "-x,v", "-i", "-U", "-q", '-t-"Gramps database"']
    ci_cmd = ["ci", "-x,v", "-q", "-f"]
    archive_name = filename + ",v"

    glade = Glade(toplevel="comment")
    top = glade.toplevel
    text = glade.get_object("description")
    top.set_transient_for(parent)
    top.run()
    comment = text.get_text()
    top.destroy()

    if not os.path.isfile(archive_name):
        cmd = init + [archive_name]
        proc = subprocess.Popen(cmd, stderr=subprocess.PIPE)
        status = proc.wait()
        message = "\n".join(proc.stderr.readlines())
        proc.stderr.close()
        del proc

        if status != 0:
            ErrorDialog(
                _("Archiving failed"),
                _(
                    "An attempt to create the archive failed "
                    "with the following message:\n\n%s"
                )
                % message,
                parent=self.top,
            )

    if cursor_func:
        cursor_func(_("Creating data to be archived..."))

    plugin_manager = GuiPluginManager.get_instance()
    for plugin in plugin_manager.get_export_plugins():
        if plugin.get_extension() == "gramps":
            export_function = plugin.get_export_function()
            export_function(dbase, filename, user)

    if cursor_func:
        cursor_func(_("Saving archive..."))

    cmd = ci_cmd + ["-m%s" % comment, filename, archive_name]
    proc = subprocess.Popen(cmd, stderr=subprocess.PIPE)

    status = proc.wait()
    message = "\n".join(proc.stderr.readlines())
    proc.stderr.close()
    del proc

    if status != 0:
        ErrorDialog(
            _("Archiving failed"),
            _(
                "An attempt to archive the data failed "
                "with the following message:\n\n%s"
            )
            % message,
            parent=self.top,
        )


def bug_fix(column, renderer, model, iter_, data):
    """
    Cell data function to set the status column.

    There is a bug in pygobject which prevents us from setting a value to
    None using the TreeModel set_value method.  Instead we set it to an empty
    string and convert it to None here.
    """
    icon_name = model.get_value(iter_, ICON_COL)
    if icon_name == "":
        icon_name = None
    renderer.set_property("icon-name", icon_name)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2002-2006  Donald N. Allingham
# Copyright (C) 2011       Doug Blank <doug.blank@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from . import Rule
from ...const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


# -------------------------------------------------------------------------
# "People marked public"
# -------------------------------------------------------------------------
class IsPublic(Rule):
    """Objects not marked private."""

    name = "Objects not marked private"
    description = "Matches objects that are not indicated as private"
    category = _("General filters")

    def apply(self, db, obj):
        return not obj.get_privacy()
